/*
 * @(#)ArrayList.java	1.36 01/12/03
 *
 * Copyright 2002 Sun Microsystems, Inc. All rights reserved.
 * SUN PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
 */

package ps2;

import java.util.*;

/**
 * Resizable-array implementation of a List.  Permits all elements,
 * including <tt>null</tt>.
 *
 * The <tt>size</tt>, <tt>isEmpty</tt>, <tt>get</tt>, 
 * and <tt>iterator</tt> methods run in constant time.  The <tt>add</tt>
 * method runs in <i>amortized constant time</i>, that is, adding n
 * elements requires O(n) time.  All of the other operations run in
 * linear time. The methods from <tt>java.util.LinkedList</tt> class 
 * from the Java 
 * library also run in linear time, but BasicList has a smaller constant 
 * factor.
 * <p/>
 * <b>Note</b>: throughout this API the notation <tt>e1.equals(e2)</tt> is used to mean
 * <tt>(e1==null ? e2==null : e1.<b>equals</b>(e2))</tt> where the second
 *  <b>equals</b> is the method that tests for object equality.
 *
 * @specfield map: (nonnegative integer -> element) // maps indexes to elements of this BasicList.
 * @derivedfield indexes: set nonnegative-integer // the keys of this.map
 * @derivedfield size: nonnegative integer // the number of elements in this BasicList.
 * 
 */

public class BasicList 
{
    /**
     * The array buffer into which the elements of the BasicList are stored.
     * The capacity of the BasicList is the length of this array buffer.
     */
    private List elementData = null;

    /**
     * Constructs an empty list with the specified initial capacity.
     *
     * @param   initialCapacity   the initial capacity of the list.
     * @exception IllegalArgumentException if the specified initial capacity
     *            is negative
     */
    private BasicList(int initialCapacity) {
      elementData = new ArrayList();
      if (initialCapacity < 0)
	throw new IllegalArgumentException("Illegal Capacity: "+
					   initialCapacity);
    }

    /**
     * returns a new, empty BasicList
     * @returns a new, empty BasicList
     */
    public BasicList() {
	this(10);
    }

    /**
     * Creates a new BasicList populated with the elements in the specified Collection.
     * 
     * returns a new BasicList | all i: [0..iterator.size) | this[i] == c.iterator[i]
     * @returns a new BasicList | all i: [0..iterator.size) | this[i] == c.iterator[i]
     * 
     * @throws NullPointerException when c == null.
     */
    public BasicList(Collection c) {
      elementData = new ArrayList(c);
    }

    /**
     * Returns the size of this list.
     * @return this.size
     */
    public int size() {
       return elementData.size();
    }

    /**
     * Tests if this list has no elements.
     *
     * @return  this.size == 0;
     */
    public boolean isEmpty() {
      return size() == 0;
    }

    /**
     * Returns <tt>true</tt> if this list contains the specified element.
     *
     * @return if some i: this.indexes | this[i].equals(elem)
     * 
     * @see Object#equals(Object)
     */
    public boolean contains(Object elem) {
	return indexOf(elem) > 0;
    }

    /**
     * Searches for the first occurence of the given argument.
     *
     * @return if some i: this.indexes | this[i].equals(elem) then
     *           returns minimum {i: this.indexes | this[i].equals(elem)}
     *         else -1
     * 
     * @see Object#equals(Object)
     */
    public int indexOf(Object elem) {
      /**	if (elem == null) {
	    for (int i = 0; i < size; i++)
		if (elementData[i]==null)
		    return i;
	} else {
      **/
      //BUG!//
      if (elem == null)
	throw new NullPointerException();
      return elementData.indexOf(elem);
    }


    /**
     * Returns a new array containing all of the elements in this list.
     *
     * @return new array: Object[ ] | array.length == this.size
     *   and all i : this.indexes | array[i] == this[i] 
     */
    public Object[] toArray() {
      /**
      Object[] result = new Object[size];
	System.arraycopy(elementData, 0, result, 0, size);
	return result;
      **/
      return elementData.toArray();
    }

    /**
     * Returns an array containing all of the elements in this list; 
     * the runtime type of the returned array is that of the
     * specified array.  If the list fits in the specified array, it is
     * returned therein.  Otherwise, a new array is allocated with the runtime
     * type of the specified array and the size of this list.<p>
     *
     * If the list fits in the specified array with room to spare (i.e., 
     * this.size < a.length), the elements in the array
     * following the end of the collection are set to
     * <tt>null</tt>.  This is useful in determining the length of the list
     * <i>only</i> if the caller knows that the list does not contain any
     * <tt>null</tt> elements.
     *
     * @modifies a.map
     * 
     * @effects if this.size &lt;= a.size then <br>
     *            all i: this.indexes | a[i] == this[i] 
     *            and all j: this.size &lt;= j &lt; a.size | 
     *            a[j] == null
     *            <br>
     *          else <br>
     *            all i: this.indexes | b[i] == this[i]
     * @return if this.size &lt;= a.length returns a <br>
     *         else returns b: Object[ ] |
     * b.length == this.size and all i: this.indexes | 
     * b[i] == this[i] 
     * @throws IllegalArgumentException when a==null.
     *         ArrayStoreException when the runtime type of a is not a supertype
     *         of the runtime type of every non-null element in this list.
     *         
     */
    public Object[] toArray(Object a[]) {
      //        if (a.length < size)  //BUG//
        if(a==null) throw new IllegalArgumentException();
      a = (Object[])java.lang.reflect.Array.newInstance(
							a.getClass().getComponentType(), size());
      
	System.arraycopy(elementData.toArray(), 0, a, 0, size());

        if (a.length > size())
            a[size()] = null;

        return a;
    }

    // Positional Access Operations

    /**
     * Returns the element at the specified position in this list.
     *
     * @return this[index]
     * @throws    IndexOutOfBoundsException when index
     * 		  &lt; 0 || index &gt;= this.size
     */
    public Object get(int index) {
      return elementData.get(index);
    }

    /**
     * Appends the specified element to the end of this list.
     *
     * @modifies this.map
     * @effects this.size == \old(this.size) + 1 and this[this.size-1] == o
     *            and all i: \old (this.indexes) | this[i] = \old(this[i])
     * @return <tt>true</tt>
     */
    public boolean add(Object o) {
      return elementData.add(o);
    }

    /**
     * Inserts the specified element at the specified position in this
     * list. Shifts the element currently at that position (if any) and
     * any subsequent elements to the right (adds one to their indices).
     *
     * @modifies this.map
     * @effects this.size == \old(this.size) + 1 <br>
     *          all i: 0 &lt;= i &lt; index | this[i] == \old(this[i])  <br>
     *          this[index] == element <br>
     *          all i: index &lt; i &lt; this.size | this[i] == \old(this[i-1]) 
     * @throws    IndexOutOfBoundsException when  index &lt; 0 || index &gt; this.size
     */
    public void add(int index, Object element) {
      elementData.add(index, element);
    }

    /**
     * Removes the element at the specified position in this list.
     * Shifts any subsequent elements to the left (subtracts one from their
     * indices).
     *
     * @modifies this.map
     * @return \old(this[index])
     * @effects this.size == \old(this.size) - 1 <br>
     *          all i: 0 &lt;= i &lt; index | this[i] == \old(this[i])  <br>
     *          all i: index &lt;= i &lt; this.size | this[i] == \old(this[i+1]) 
     * @throws    IndexOutOfBoundsException when index
     * 		  &lt; 0 || index &gt;= this.size
     */
    public Object remove(int index) {
      elementData.remove(index);
      //BUG//
      return null;
    }

    /**
     * Returns an iterator over the elements in this list in proper
     * sequence. <p>
     * 
     * @return iterator: all i: this.indexes | this[i] == iterator[i] and this.size == iterator.size
     */
    public Iterator iterator() {
      return elementData.iterator();
    }

    // Comparison and hashing

    /**
     * Compares the specified object with this list for equality.  Returns
     * <tt>true</tt> if and only if the specified object is also a list, both
     * lists have the same size, and all corresponding pairs of elements in
     * the two lists are <i>equal</i>. In other words, two lists are defined to be
     * equal if they contain the same elements in the same order.<p>
     * 
     * @return if o instanceof BasicList and <br>
     * this.size == o.size and all i: this.indexes | this[i].equals(o[i])
     */
    public boolean equals(Object o) {
	if (o == this)
	    return true;
	if (!(o instanceof BasicList))
	    return false;

	Iterator e1 = iterator();
	Iterator e2 = ((BasicList) o).iterator();
	while(e1.hasNext() && e2.hasNext()) {
	    Object o1 = e1.next();
	    Object o2 = e2.next();
	    if (!(o1==null ? o2==null : o1.equals(o2)))
		return false;
	}
	//	return !(e1.hasNext() || e2.hasNext());
	//BUG!!//
	return true;
    }

    /**
     * Returns the hash code value for this list. <p>
     *
     * This implementation uses exactly the code that is used to define the
     * list hash function in the documentation for the <tt>List.hashCode</tt>
     * method.
     * 
     * @return the hash code value for this list.
     * @see     List#hashCode()
     */
    public int hashCode() {
	int hashCode = 1;
	Iterator i = iterator();
     	while (i.hasNext()) {
	    Object obj = i.next();
	    hashCode = 31*hashCode + (obj==null ? 0 : obj.hashCode());
	}
	return hashCode;
    }

}
