package ps2;

import junit.framework.*;
import java.util.List;
import java.util.ArrayList;
import java.util.Iterator;

/** This class contains a set of test cases that can be used to 
    test the implementation of the BasicList class.
    <p>
*/
public class BasicListTest extends TestCase {
  
  BasicList emptyList = new BasicList();
  BasicList eltList = new BasicList();

  public void setUp() {
    //this method can be used to initialize variables that can be
    //re-used for other tests
    //emptyList = new BasicList();
  }
  
  
  public void testEmptyConstructor() {
    BasicList bl = new BasicList();
    assertNotNull("basicList should != null after call to constructor", bl );
    assertEquals("Size on basicList should return 0", 0, bl.size());
  }
  
  public void testCollectionConstructor() {
    List elts = null;
    
    //check to see that this throws NPE
    boolean threwNPE = false;
    try {
      eltList = new BasicList( elts );
    }
    catch(Exception e) {
      threwNPE = true;
    }
    assertTrue( "Constructor should throw exception if given null collection",threwNPE );
    
    elts = new ArrayList();
    elts.add("hi");
    elts.add("hello");
    elts.add("what");
   
    //check order...
    eltList = new BasicList( elts );
    assertNotNull("eltList should != null after call to constructor", eltList );
    assertEquals("Size should return 3", 3, eltList.size());
    assertEquals("The first element returned should be \"hi\"", "hi", (String)eltList.get(0));
    assertEquals("The second element returned should be \"hello\"", "hello", (String)eltList.get(1));
    assertEquals("The third element returned should be \"what\"", "what", (String)eltList.get(2));
  }

  
  public void testSize() {
    assertEquals("Empty List should return size 0", 0, emptyList.size());
        
    eltList.add("hi");
    eltList.add("hello");
    eltList.add("what");
    
    assertEquals("Element list should return size 3", 3, eltList.size());
  }
  
  public void testIsEmpty() {
    assertTrue("Empty list is empty", emptyList.isEmpty());
    eltList.add("hi");
    eltList.add("hello");
    assertFalse("Element list is not empty", eltList.isEmpty());
  }

  public void testContains() {
    assertFalse("Empty list does not contain anything", emptyList.contains("hi"));
    
    eltList.add( "hi" );
    //NIS:BUG HERE!!
    assertTrue("EltList should now have hi", eltList.contains("hi"));
    assertFalse("EltList should not have hello", eltList.contains("hello"));
    eltList.add("hello" );
    assertTrue("EltList should have hello", eltList.contains("hello"));
  }

  public void testIndexOf() {
    assertEquals("Index of \"hi\" in empty list should be -1", -1, emptyList.indexOf("hi"));
    //NIS:BUG HERE!!
    assertEquals("Index of null in empty list should be -1", -1, emptyList.indexOf(null));
    
    emptyList.add("hi");
    assertEquals("Index of \"hi\" in emptyList should be 0", 0, emptyList.indexOf("hi"));
    
    emptyList.add("hi");
    assertEquals("Index of \"hi\" in emptyList should be 0", 0, emptyList.indexOf("hi"));
  }
  
  public void testEquals() {
    //test that null objects and objects of other run-time types
    //return false
    
    assertTrue( "List should be equal to itself",emptyList.equals(emptyList));
    assertTrue("comparing two empty list should return true", emptyList.equals(new BasicList()));
    assertFalse("list != null",emptyList.equals(null));
    assertFalse( "list != not list",emptyList.equals("hi") );
    
    eltList.add("hi");
    assertFalse( "comparing list with differnt size", emptyList.equals(eltList) );
    
    emptyList.add("hi");
    //NIS: Bug here!!!
    assertTrue( "comparing lists with same size and elements",emptyList.equals(eltList));
    emptyList.add(null);
    assertFalse( "comparing list with differnt elts", emptyList.equals(eltList) );
    
    eltList.add( "hello" );
    eltList.add( "nis" );
    assertFalse( "comparing list with differnt elts", emptyList.equals(eltList) );
    
  }
  
  public void testToArray() {
    Object[] arr = emptyList.toArray();
    assertEquals("Empty List to array should give empty array", 0, arr.length);

    emptyList.add("hi");
    emptyList.add("hello");
    emptyList.add("what");
    emptyList.add("hi");
    arr = emptyList.toArray();
    assertEquals("empty List now has 4 elements", 4, arr.length);
    assertEquals("first element in array should be \"hi\"","hi",(String)arr[0]);
    assertEquals("Second element in array should be \"hello\"","hello",(String)arr[1]);
    assertEquals("Third element in array should be \"what\"","what",(String)arr[2]);
    assertEquals("fourth element in array should be \"hi\"","hi",(String)arr[3]);
  }


  public void testToArrayWithArg() {
    Integer zero = new Integer(0);
    Integer one = new Integer(1);
    Integer two = new Integer(2);
    Integer three = new Integer(3);
    
    Object[] a = { one, two };
    boolean threwASE = false;

    emptyList.add(zero);
    Object[] b = emptyList.toArray( a );
    
    //NIS:BUG HERE!!
    assertTrue( "a and b should point to same memory", a == b );
    assertEquals( "a and b should have same elements", a[0], b[0] );
    assertEquals( "a and b should have same elements", a[1], b[1] );
    assertTrue( "a and b should have same length", a.length == b.length );
    assertTrue( "First spare index in a should be null", a[1] == null );
    assertEquals( "List does not have null, lenght array = length list", 1, a.length );
    a[0] = three;
    assertEquals( "b[0] should be three as well", three, (Integer)b[0] );

    
    emptyList.add(one);
    emptyList.add(two);
    b = emptyList.toArray();
    assertFalse( "a and b should be different", a[0] == b[0] );
    assertFalse( "a and b have different lengths", a.length == b.length );

    //check order and all elements
    assertEquals( "First element should be Integer 0", zero, (Integer)b[0] );
    assertEquals( "Second element should be Integer 1", one, (Integer)b[1] );
    assertEquals( "Third element should be Integer 2", two, (Integer)b[2] );

    a = new String[2];
    try {
      emptyList.toArray( a );
    }
    catch( Exception e ) {
      threwASE = true;
    }
    assertTrue("passing array with type mismatch should throw exception", threwASE );    
  }

  public void testGet( ) {
    boolean threwIOOBE = false;
    try{
      Object o = emptyList.get(-1);
    }
    catch( Exception e ) {
      threwIOOBE = true;
    }
    assertTrue( "calling get with -ve index should throw", threwIOOBE );

    threwIOOBE = false;
    try{
      Object o = emptyList.get(1);
    }
    catch( Exception e ) {
      threwIOOBE = true;
    }
    assertTrue( "calling get on empty list with +ve index should throw", threwIOOBE );

    emptyList.add( "hi" );
    emptyList.add( "hello" );
    assertEquals( "Element at index 1 should be \"hello\"", "hello", (String) emptyList.get(1) );
  }

  //read spec for collection.add
  public void testAdd( ) {
    assertTrue( "Empty list is empty", emptyList.isEmpty() );
    
    boolean ret = emptyList.add( "hi" );
    assertEquals("list size should be 1", 1, emptyList.size());
    assertTrue( "add should return true", ret );
    
    //NIS: Commenting this line b/c this is a spurious error. Really it is
    //contains(elem) that is buggy and not add.

    //assertTrue( "empty list should now have \"hi\"", emptyList.contains("hi"));
    assertEquals( "Element at index 0 should be hi", "hi", (String)emptyList.get(0));
    
    assertTrue( "add should return true", ret );
    
    ret = emptyList.add("hello");
    assertTrue( "empty list should now have \"hello\"", emptyList.contains("hello"));
    assertEquals( "Element at index 0 should be hi", "hi", (String)emptyList.get(0));
    assertEquals( "Element at index 1 should be hello", "hello", (String)emptyList.get(1));
    assertEquals("list size should be 2", 2, emptyList.size());
    assertTrue( "add should return true", ret );

    ret = emptyList.add( "hi" );
    assertTrue( "add should return true", ret );
  }

  public void testAddWithArgs() {
    boolean threwIOOBE = false;

    try {
      emptyList.add( -1, "hi" );
    }
    catch( Exception e ) {
      threwIOOBE = true;
    }
    assertTrue( "Throw exception on adding to -ve index", threwIOOBE );

    threwIOOBE = false;
    try {
      emptyList.add( 1, "hi" );
    }
    catch( Exception e ) {
      threwIOOBE = true;
    }
    assertTrue( "Throw exception on adding to index > size", threwIOOBE );
    
    emptyList.add(0,"hi");
    assertEquals("List should have hi at position 0", "hi", (String)emptyList.get(0) );
    assertEquals("List size should be 1", 1, emptyList.size());

    emptyList.add( "hello" );
    emptyList.add( "hiya" );
    emptyList.add( 1, "nis" );

    assertEquals( "Size should be 4", 4, emptyList.size() );
    assertEquals( "Index of hi should be 0", 0, emptyList.indexOf( "hi" ) );
    assertEquals( "Index of nis should be 1", 1, emptyList.indexOf( "nis") );
    assertEquals( "Index of hello should be 2", 2, emptyList.indexOf( "hello" ));
    assertEquals( "Index of hiya should be 3", 3, emptyList.indexOf( "hiya" ) );
  }


  public void testRemove() {
    boolean threwIOOBE = false;

    try {
      emptyList.remove( -1 );
    }
    catch( Exception e ) {
      threwIOOBE = true;
    }
    assertTrue( "Throw exception on adding to -ve index", threwIOOBE );

    threwIOOBE = false;
    try {
      emptyList.remove( 1 );
    }
    catch( Exception e ) {
      threwIOOBE = true;
    }
    assertTrue( "Throw exception on adding to index > size", threwIOOBE );

    threwIOOBE = false;
    try {
      emptyList.remove( 0 );
    }
    catch( Exception e ) {
      threwIOOBE = true;
    }
    assertTrue( "Throw exception on adding to index > size", threwIOOBE );

    emptyList.add( "hi" );
    emptyList.add( "hello" );
    assertEquals("size of list is 2", 2, emptyList.size() );

    Object ret = emptyList.remove(1);
    //NIS:BUG HERE!!
    assertEquals( "removed object should be \"hello\"", "hello", (String)ret );
    assertFalse( "list should not contail \"hello\"", emptyList.contains("hello"));
    assertEquals( "List size should be 1", 1, emptyList.size());
  }

  public void testIterator() {
    emptyList.add( "hi" );
    emptyList.add( "nis" );
    emptyList.add( "hello" );

    Iterator iter = emptyList.iterator();
    int cnt = 0;
    while( iter.hasNext() ) {
      cnt++;
      iter.next();
    }
    
    iter = emptyList.iterator();
    assertEquals( "iterator should return size number of elements", 3, cnt );
    assertEquals( "First iter element should be hi", "hi", (String)iter.next() );
    assertEquals( "First iter element should be nis", "nis", (String)iter.next() );
    assertEquals( "First iter element should be hello", "hello", (String)iter.next() );
  }

  public void testHashCode() {
    assertEquals( "equal lists should have same hashcode", eltList.hashCode(), emptyList.hashCode() );
    emptyList.add( "hi" );
    assertTrue( "unequal lists should have different hashcode", eltList.hashCode() != emptyList.hashCode() );
    eltList.add( "hi" );
    assertEquals( "equal lists should have same hashcode", eltList.hashCode(), emptyList.hashCode() );
    assertTrue( "unequal objects should have different hashcode", eltList.hashCode() != "hi".hashCode() );
    emptyList.add(null);
    eltList.add(null);
    assertEquals( "equal lists should have same hashcode", eltList.hashCode(), emptyList.hashCode() );
  }
  
  //
  // INCLUDE OTHER TEST METHODS HERE!!! (and, optionally, add method
  // names to suite() method below)
  //
  
  public BasicListTest(String name) {
    super(name);
  }
  
  // Tell JUnit what order to run the tests in
  
  //NIS: Addign in no particular order here
  public static Test suite() {
    TestSuite suite = new TestSuite();
    
    suite.addTest(new BasicListTest("testEquals"));
    suite.addTest(new BasicListTest("testEmptyConstructor"));
    suite.addTest(new BasicListTest("testCollectionConstructor"));
    suite.addTest(new BasicListTest("testSize"));
    suite.addTest(new BasicListTest("testIsEmpty"));
    suite.addTest(new BasicListTest("testContains"));
    suite.addTest(new BasicListTest("testIndexOf"));
    suite.addTest(new BasicListTest("testToArray"));
    suite.addTest(new BasicListTest("testToArrayWithArg"));
    suite.addTest(new BasicListTest("testGet"));
    suite.addTest(new BasicListTest("testAdd"));
    suite.addTest(new BasicListTest("testAddWithArgs"));
    suite.addTest(new BasicListTest("testRemove"));
    suite.addTest(new BasicListTest("testIterator"));
    suite.addTest(new BasicListTest("testHashCode"));
    
    return suite;
  }
  
}
